//
// Use at your own risk. But if you find it usefull and would like to use it
//
#include "stdafx.h"
#include "resource.h"
#include <stdio.h>


CComModule _Module;

#define moduleInstance			_Module.GetModuleInstance()
#define moduleResourceInstance	_Module.GetResourceInstance()

#define DEFAULT_BUFFER_SIZE		1024
#define MAX_BUFFER_SIZE			4096


typedef CString CMyString;
CSimpleArray<CMyString>			arr;

// special URL Commands
// http://<HOST>/reset		Update parameters from 'config.ini'
// http://<HOST>/uselog		Start writing rewrites to log file
// http://<HOST>/skiplog	Stop writing rewrites to log file
CHAR CMD_RESET[]			= "/reset";
CHAR CMD_USE_LOG[]			= "/uselog";
CHAR CMD_SKIP_LOG[]			= "/skiplog";
CHAR CMD_LOG_STATUS[]		= "/logstatus";
CHAR CMD_ABOUT[]			= "/about";
CHAR CONFIG_INI[]			= "config.ini";
//
// HTML Pages definitions for special commands
//
CHAR HTML_INITIALIZED[]		= "initialized.html";
CHAR HTML_LOG_SET[]			= "logSet.html";
CHAR HTML_LOG_UNSET[]		= "logUnSet.html";
CHAR HTML_ABOUT[]			= "about.html";

CHAR  szUrl[]				= "url";

//
// Various globals
//
HANDLE						g_hLogFile = INVALID_HANDLE_VALUE;
CRITICAL_SECTION			g_LogFileLock;
CRITICAL_SECTION			g_FileLock;
BOOL						g_Log;

//
// Filter Required functions
//
BOOL WINAPI		GetFilterVersion(HTTP_FILTER_VERSION *pVer);
DWORD WINAPI	HttpFilterProc(IN HTTP_FILTER_CONTEXT *pfc, DWORD dwNotificationType, LPVOID pvNotification);
DWORD			OnPreprocHeaders(HTTP_FILTER_CONTEXT *pfc, HTTP_FILTER_PREPROC_HEADERS *pPPH);
BOOL			WINAPI TerminateFilter(DWORD dwFlags);


void			WriteDebugString(LPCSTR string);
bool			DoInitialize(HMODULE hModule);

void			DoReset(HTTP_FILTER_CONTEXT *pfc, HTTP_FILTER_PREPROC_HEADERS *pPPH, HMODULE hModule);
CMyString		DoReplace(CMyString urlString);


bool			IsSpecialCommand(HTTP_FILTER_CONTEXT *pfc, HTTP_FILTER_PREPROC_HEADERS *pPPH, LPCSTR string, DWORD* dwRet);
void			setUseLog(HTTP_FILTER_CONTEXT *pfc, BOOL bSet);
void			showWebPage(HTTP_FILTER_CONTEXT *pfc, LPSTR szHTMLResouce);


//
// returns if we want to write to the log file or not
//
inline BOOL getUseLog()
{
	return g_Log;
}

//
// Shows the web page related to command from resources
//
void showWebPage(HTTP_FILTER_CONTEXT *pfc, LPSTR szHTMLResouce)
{
	HRSRC hRes = ::FindResource(moduleResourceInstance, szHTMLResouce, RT_HTML);
	HGLOBAL hGlobal = LoadResource(moduleResourceInstance, hRes);
	LPVOID htmlData = LockResource(hGlobal);
	DWORD dwSize = SizeofResource(moduleResourceInstance, hRes);

	pfc->WriteClient(pfc, htmlData, &dwSize, NULL);

	UnlockResource(hGlobal);
	FreeResource(hGlobal);
}

//
// sets if we want to write to the log file or not.
// also redirects to appropriate web page (http://<HOST>/logSet.html or http://<HOST>/logUnSet.html)
// 
void setUseLog(HTTP_FILTER_CONTEXT *pfc, BOOL bSet)
{
	g_Log = bSet;
	if (g_Log)
		showWebPage(pfc, HTML_LOG_SET);
	else
		showWebPage(pfc, HTML_LOG_UNSET);
}

//
//Writes to the log file (if we want to use it of course)
//
void WriteDebugString(LPCSTR string)
{
	DEBUG_WRITELINE( string );
	if (!getUseLog())
		return;

	DWORD cbBuf = DEFAULT_BUFFER_SIZE;
	BOOL fRet = FALSE;

	EnterCriticalSection(&g_LogFileLock);

	SetFilePointer(g_hLogFile, 0, NULL, FILE_END);
	fRet = WriteFile(g_hLogFile, string, lstrlen(string), &cbBuf, NULL);
	fRet = WriteFile(g_hLogFile, "\n", 1, &cbBuf, NULL);

	LeaveCriticalSection(&g_LogFileLock);
}

//
//Checks if the url provided is actually one of the special commands
//and executes the appropriate function.
//
bool IsSpecialCommand(HTTP_FILTER_CONTEXT *pfc, HTTP_FILTER_PREPROC_HEADERS *pPPH, LPCSTR string, DWORD* dwRet)
{
	if (!lstrcmp(string, CMD_RESET))
	{
		DoReset(pfc, pPPH, moduleInstance);
	}
	else if (!lstrcmp(string, CMD_USE_LOG))
	{
		setUseLog(pfc, TRUE);
	}
	else if (!lstrcmp(string, CMD_SKIP_LOG))
	{
		setUseLog(pfc, FALSE);
	}
	else if (!lstrcmp(string, CMD_LOG_STATUS))
	{
		if (getUseLog())
			showWebPage(pfc, HTML_LOG_SET);
		else
			showWebPage(pfc, HTML_LOG_UNSET);
	}
	else if (!lstrcmp(string, CMD_ABOUT))
	{
		showWebPage(pfc, HTML_ABOUT);
	}
	else 
	{
		return false;
	}

	*dwRet = SF_STATUS_REQ_FINISHED;
	return true;
}

//
//Initialize the replacements array from the configuration file.
//
bool DoInitialize(HMODULE hModule)
{
	LPSTR szFileName = new CHAR[MAX_PATH];

    if (0 != ::GetModuleFileName(hModule, szFileName, MAX_PATH))
    {
        ::PathRemoveFileSpec(szFileName);
        ::PathAppend(szFileName, CONFIG_INI);
		LPSTR tmpStr = new CHAR[DEFAULT_BUFFER_SIZE];

	    EnterCriticalSection(&g_FileLock);

        FILE* hFile = fopen(szFileName, "r");

		CMyString str;
		arr.RemoveAll();
        while (fgets(tmpStr, MAX_PATH, hFile) != NULL)
        {
            StrTrimA(tmpStr, " \r\n\t");
			str = tmpStr;
			arr.Add(str);
        }
		fclose(hFile);
		delete [] tmpStr;
		LeaveCriticalSection(&g_FileLock);
	}
	delete [] szFileName;
	return true;
}

//
// handles the http://<HOST>/reset command.
// resets the replacements array to the values in config.ini
//
void DoReset(HTTP_FILTER_CONTEXT *pfc, HTTP_FILTER_PREPROC_HEADERS *pPPH, HMODULE hModule)
{
	DEBUG_WRITELINE( "DoReset" );

	DoInitialize(hModule);
	pPPH->SetHeader(pfc, szUrl, HTML_INITIALIZED);
}

//
// Uses the replacements array to replace all parameters from the url.
//
CMyString DoReplace(CMyString urlString)
{
	CMyString urlAddress, urlParams;

	WriteDebugString( urlString );

	//split the url address to address and parameters 
	int qPos = urlString.Find("?");
	if (qPos == -1)
	{
		urlAddress = urlString;
	}
	else
	{
		urlAddress = urlString.Left(qPos);
		urlParams = urlString.Mid(urlAddress.GetLength()+1);
	}

	//do replacements.
	int i;
	for (i = 0; i < arr.GetSize(); i += 2)
	{
		CMyString& szSearch = arr[i];
		int iFind = urlAddress.Find(szSearch);
		if (iFind >= 0)
		{
			urlAddress.Replace(szSearch,arr[i+1]);
		}
	}

	//combine address and params to a full url address
	urlString = urlAddress;
	urlString += "&";
	urlString += urlParams;

	//change the first '&' to a '?' to mark start of parameters 
	qPos = urlString.Find('&');
	if (qPos > -1)
		urlString.SetAt(qPos, '?');

	//trim final ? or &
	i = urlString.GetLength();
	if (i > 0)
	{
		CHAR ch = urlString.GetAt(i-1);
		switch (ch)
		{
		case '?':
		case '&':
			urlString.SetAt(i-1, 0);
			break;
		default:
			break;
		}
	}
	//if url starts with ? add an initial /
	if (qPos == 0)
	{
		urlString = "/" + urlString;
	}

	WriteDebugString( urlString );

	return urlString;
}

//
// Handle the `SF_NOTIFY_PREPROC_HEADERS` notification
//
DWORD OnPreprocHeaders(HTTP_FILTER_CONTEXT *pfc, HTTP_FILTER_PREPROC_HEADERS *pPPH)
{
    DWORD	dwRet	= SF_STATUS_REQ_NEXT_NOTIFICATION;
    BOOL	fRet	= FALSE;
    CHAR	*pBuf	= new CHAR[DEFAULT_BUFFER_SIZE];
    CHAR	*pszBuf	= pBuf;
    DWORD	cbBuf	= DEFAULT_BUFFER_SIZE;

    SetLastError(NO_ERROR);

    if (pfc == NULL || pPPH == NULL)
    {
        SetLastError(ERROR_INVALID_PARAMETER);
        goto Finished;
    }

    if (fRet == FALSE)
    {
        pszBuf = new CHAR[cbBuf];
        if (pszBuf == NULL)
        {
            SetLastError(ERROR_NOT_ENOUGH_MEMORY);
            goto Finished;
        }

		fRet = pPPH->GetHeader(pfc, szUrl, pszBuf, &cbBuf);
		if (fRet == FALSE)
		{
			goto Finished;
		}
		//
		// At this point, pszBuf points to the raw URL...
		//
		if (!IsSpecialCommand(pfc, pPPH, pszBuf, &dwRet))
		{
			CMyString newUrl = DoReplace( T2A(pszBuf) );
			pPPH->SetHeader(pfc, szUrl, (LPSTR)(LPCSTR)newUrl);
		}
    }

    SetLastError(NO_ERROR);

Finished:

    if (pszBuf != pBuf)
    {
        delete pszBuf;
    }

    if (GetLastError() != NO_ERROR)
    {
        dwRet = SF_STATUS_REQ_ERROR;
    }
	delete [] pBuf;

    return dwRet;
}

//
// Required by ISAPI EXTENSION
//
extern "C" BOOL WINAPI GetFilterVersion(HTTP_FILTER_VERSION *pVer)
{
    InitializeCriticalSection(&g_LogFileLock);
    pVer->dwFilterVersion = HTTP_FILTER_REVISION;

	//Set the extension name to the value in our resources' string table
	::LoadStringA(moduleResourceInstance, IDS_FILTER_DESCRIPTION, pVer->lpszFilterDesc, SF_MAX_FILTER_DESC_LEN);

	pVer->dwFlags = SF_NOTIFY_ORDER_HIGH | SF_NOTIFY_PREPROC_HEADERS ;

	//
    // Initialize the logging critical section
    //
    if (!DoInitialize(_Module.GetModuleInstance()))
    {
        return FALSE;
    }

	LPSTR szFileName = new CHAR[MAX_PATH];

    ::GetModuleFileName(_Module.GetModuleInstance(), szFileName, MAX_PATH);
    wsprintf(szFileName, "%s.%d.log", szFileName, GetCurrentProcessId());

	g_hLogFile = CreateFile(
        szFileName,
        GENERIC_WRITE,
        FILE_SHARE_READ | FILE_SHARE_WRITE,
        NULL,
        OPEN_ALWAYS,
        FILE_ATTRIBUTE_NORMAL,
        NULL);

	delete [] szFileName;

	if (g_hLogFile == INVALID_HANDLE_VALUE)
    {
        return FALSE;
    }

	g_Log = TRUE;

	WriteDebugString( pVer->lpszFilterDesc );
	WriteDebugString( "*** Log File Initialized ***" );

	g_Log = FALSE;

    return TRUE;
}

// Required by ISAPI EXTENSION
extern "C" DWORD WINAPI HttpFilterProc(IN HTTP_FILTER_CONTEXT *pfc, DWORD dwNotificationType, LPVOID pvNotification)
{
    switch (dwNotificationType)
    {
    case SF_NOTIFY_PREPROC_HEADERS:
        return OnPreprocHeaders(pfc, (HTTP_FILTER_PREPROC_HEADERS *) pvNotification);
    }

    return SF_STATUS_REQ_NEXT_NOTIFICATION;
}

//
// Called when IIS terminates ISAPI EXTENSION. Perform some clean-up.
//
extern "C" BOOL WINAPI TerminateFilter(DWORD dwFlags)
{
	dwFlags;

	DEBUG_WRITE( "TerminateFilter" );
	arr.RemoveAll();

    if (g_hLogFile != INVALID_HANDLE_VALUE)
    {
        CloseHandle(g_hLogFile);
        g_hLogFile = INVALID_HANDLE_VALUE;
    }

    DeleteCriticalSection(&g_LogFileLock);

    return TRUE;
}
